import os, requests, time


class SWAuth(object):
    """
    SWAuth allows for a user to authenticate utilizing
    their standard Strangeworks login via Auth0
    """

    def __init__(
        self,
        auth0_url: str = "https://auth.quantumcomputing.com",
        auth0_client_id: str = "TZN22C2vD1OW8pPPAomhgWqiKxpgUKnJ",
        auth0_audience: str = "https://strangeworks.com/login",
        timeout_secs: int = 300,
        frequency_secs: int = 5,
    ):
        self.auth0_url = (
            os.getenv("STRANGEWORKS_AUTH_URL", default=None)
            if auth0_url is None
            else auth0_url
        )
        self.auth0_client_id = (
            os.getenv("STRANGEWORKS_AUTH_CLIENT_ID", default=None)
            if auth0_client_id is None
            else auth0_client_id
        )
        self.auth0_audience = (
            os.getenv("STRANGEWORKS_AUTH_AUDIENCE", default=None)
            if auth0_audience is None
            else auth0_audience
        )
        self.timeout_secs = (
            os.getenv("STRANGEWORKS_AUTH_TIMEOUT_SECS", default=None)
            if timeout_secs is None
            else timeout_secs
        )
        self.frequency_secs = (
            os.getenv("STRANGEWORKS_AUTH_FREQUENCY", default=None)
            if frequency_secs is None
            else frequency_secs
        )

    def _get_device_code(self):
        url = f"{self.auth0_url}/oauth/device/code"

        res = requests.post(
            url=url,
            data={
                "client_id": self.auth0_client_id,
                "audience": self.auth0_audience,
                "scope": "openid",
            },
            headers={"ContentType": "application/json; charset=utf-8"},
        )
        return res.json()

    def _get_token_loop(self, device_code):
        timeout = time.time() + self.timeout_secs

        while True:
            time.sleep(self.frequency_secs)
            res = self._get_token(device_code)

            if res.status_code == requests.codes.ok:
                raw = res.json()
                return (raw["access_token"], raw["id_token"])

            if time.time() > timeout:
                break

        return None

    def _get_token(self, device_code):
        url = f"{self.auth0_url}/oauth/token"
        res = requests.post(
            url=url,
            data={
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                "device_code": device_code,
                "client_id": self.auth0_client_id,
            },
            headers={"ContentType": "application/json; charset=utf-8"},
        )
        return res


def Login():
    c = SWAuth()
    dc = c._get_device_code()
    print(
        f"Complete authentication by visiting this url: {dc['verification_uri_complete']}"
    )
    return c._get_token_loop(dc["device_code"])



import os, yaml

from strangeworks.annealing.annealing import Annealing
from strangeworks.circuit_runner.circuit_runner import CircuitRunner


from . import __version__
from strangeworks.rest_client.rest_client import StrangeworksRestClient


class Client(object):
    """Strangeworks client implements the Strangeworks API and provides core functionality for cross-vendor applications"""

    def __init__(
        self,
        username: str = None,  # the unique username found on quantumcomputing.com
        api_key: str = None,  # a secret api_key that can be accessed on quantumcomputing.com
        auth_token: str = None,  # a jwt token used for authorization.
        host: str = None,  # the host URL. defaults to https://api.quantumcomputing.com
        config_location: str = None,  # the config location to the SDK configuration. can ge generated using the Strangeworks CLI
        auth_location: str = None,  # the location of the authentication configuration. can be generated using the Strangeworks CLI
        project_id: str = None,  # the project ID to utilize for this session.
        result_id: str = None,  # the result ID to append result data to in this session
        headers: dict = None,  # headers that are sent as part of the request to Strangeworks
        rest_client: StrangeworksRestClient = None,
    ):
        """
        Client configuration is heirchical and flows as follows:
        - Configuration file
        - Environment variables
        - __init__ parameters
        """

        # set each parameter utilizing the parameter passed into the init function
        # or the environment variable in that order of preference
        self.__username = (
            os.getenv("STRANGEWORKS_USERNAME", default=None)
            if username is None
            else username
        )

        self.__api_key = (
            os.getenv("STRANGEWORKS_API_KEY", default=None)
            if api_key is None
            else api_key
        )

        self.__auth_token = (
            os.getenv("STRANGEWORKS_AUTH_TOKEN", default=None)
            if auth_token is None
            else auth_token
        )

        self.__host = (
            os.getenv("STRANGEWORKS_HOST", default=None) if host is None else host
        )

        self.__config_location = (
            os.getenv("STRANGEWORKS_CONFIG_LOCATION", default=None)
            if config_location is None
            else config_location
        )

        self.__auth_location = (
            os.getenv("STRANGEWORKS_AUTH_LOCATION", default=None)
            if auth_location is None
            else auth_location
        )

        self.__project_id = (
            os.getenv("STRANGEWORKS_PROJECT_ID", default=None)
            if project_id is None
            else project_id
        )

        self.headers = (
            os.getenv("STRANGEWORKS_HEADERS", default=None)
            if headers is None
            else headers
        )

        self.result_id = (
            os.getenv("STRANGEWORKS_RESULT_ID", default=None)
            if result_id is None
            else result_id
        )

        # set the default config location if none was set
        if self.__config_location is None:
            strange_config_directory = os.path.expanduser("~/.config/strangeworks")
            default_strange_config_file_path = f"{strange_config_directory}/config.yaml"
            if os.path.isfile(default_strange_config_file_path):
                self.config_location = default_strange_config_file_path

            # do the same for auth file ...
            auth_file_path = f"{strange_config_directory}/auth.yaml"
            if os.path.isfile(auth_file_path):
                self.__auth_location = auth_file_path

        # if a file was set load the client up utilizing the configuration file
        if self.__config_location is not None:
            with open(self.config_location, "r") as conf_file:
                cfg = yaml.safe_load(conf_file)

            self.__project_id = (
                cfg.get("project_id", None)
                if self.__project_id is None
                else self.__project_id
            )

            self.headers = (
                cfg.get("headers", None) if self.headers is None else self.headers
            )
            self.__host = cfg.get("host", None) if self.__host is None else self.__host

        # set up auth config from auth.yaml if it exists. will only
        # set value if its currently set to None
        if self.__auth_location is not None:
            with open(self.auth_location, "r") as auth_file:
                auth_cfg = yaml.safe_load(auth_file)

            self.__api_key = (
                auth_cfg.get("api_key", None) if api_key is None else api_key
            )

            self.__auth_token = (
                auth_cfg.get("auth_token", None) if auth_token is None else auth_token
            )

            self.__username = (
                auth_cfg.get("username", None) if username is None else username
            )

        if rest_client is not None:
            self.rest_client = rest_client
        else:
            self.rest_client = StrangeworksRestClient(
                headers=headers,
                host=host,
                username=username,
                api_key=api_key,
                auth_token=auth_token,
                version=__version__,
            )

        self.annealing = Annealing(
            rest_client=self.rest_client, result_id=self.result_id
        )
        self.circuit_runner = CircuitRunner(
            rest_client=self.rest_client, result_id=self.result_id
        )

    def authenticate(self, username: str = None, api_key: str = None):
        """
        authenticate is used to generate an auth token and utilized within the session
        with the username and api_key. Will overwrite an auth token that is stored
        either in configuration or memory when called
        """
        self.__username = username
        self.__api_key = api_key
        self.rest_client.authenticate(username=username, api_key=api_key)
        self.annealing._new_client(self.rest_client)
        self.circuit_runner._new_client(self.rest_client)

    def store_configuration(
        self,
        config_path: str = None,
        store_auth: bool = True,
        store_config: bool = True,
    ):
        """
        store_configuration stores the current configuration for the app in the configuration files auth.yaml and
        config.yaml specified by either the config_path or the default configuration directory
        """
        if config_path is None:
            config_path = os.path.expanduser("~/.config/strangeworks")
        if store_auth:
            with open(f"{config_path}/auth.yaml", "w") as authfile:
                yaml.dump(
                    {
                        "api_key": self.__api_key,
                        "auth_token": self.__auth_token,
                        "username": self.__username,
                    },
                    authfile,
                )
        if store_config:
            with open(f"{config_path}/config.yaml", "w") as conffile:
                yaml.dump(
                    {
                        "project_id": self.__project_id,
                        "host": self.__host,
                    },
                    conffile,
                )

    def _host(self, new_host: str = None):
        self.__host = new_host
        rc = StrangeworksRestClient(
            headers=self.headers,
            host=new_host,
            username=self.__username,
            api_key=self.__api_key,
            auth_token=self.__auth_token,
        )
        self.rest_client = rc
        self.annealing._new_client(rc)
        self.circuit_runner._new_client(self.rest_client)



import requests
from strangeworks.errors.error import StrangeworksError


class StrangeworksRestClient:
    __default_host = "https://api.quantumcomputing.com"

    def __init__(
        self,
        headers: dict = None,
        host: str = None,
        username: str = None,
        api_key: str = None,
        auth_token: str = None,
        session: requests.Session = None,
        default_timeout: int = None,
        version: str = "",
    ):
        self.__username = username
        self.__api_key = api_key
        self.headers = headers
        self.__auth_token = auth_token
        self.__default_timeout = default_timeout
        if headers is None:
            self.headers = {
                "X-Strangeworks-API-Version": "0",
                "X-Strangeworks-Client-ID": "strangeworks-sdk-python",
                "X-Strangeworks-Client-Version": version,
                "x-strangeworks-provider-account": "",
                "Authorization": f"Bearer {auth_token}",
            }

        self.__host = host
        if host is None:
            self.__host = self.__default_host

        self.__session = session
        if session is None:
            self.__session = requests.Session()

        # set up re-authentication hook!
        self.__session.hooks["response"].append(self.__reauthenticate)

        # if a user has passed in authentication vars but has not generated a token
        # do so now, otherwise pass along
        if (
            self.__auth_token is None
            and self.__username is not None
            and self.__api_key is not None
        ):
            self.authenticate(self.__username, self.__api_key)

    def post(
        self,
        url: str = "",
        json: dict = None,
        files: dict = None,
        expected_response: int = 200,
    ):
        r = self.__session.post(
            url=f"{self.__host}{url}",
            json=json,
            files=files,
            timeout=self.__default_timeout,
        )
        res = r.json()
        if r.status_code != expected_response:
            self.__parse_error(r)
        return res

    def delete(
        self,
        url: str = "",
        expected_response: int = 200,
    ):
        r = self.__session.delete(
            url=f"{self.__host}{url}",
            timeout=self.__default_timeout,
        )
        if r.status_code != expected_response:
            self.__parse_error(r)

    def get(
        self,
        url: str = "",
    ) -> dict:
        r = self.__session.get(url=f"{self.__host}{url}")
        res = r.json()
        if r.status_code != 200:
            self.__parse_error(r)
        return res

    def authenticate(self, username: str = None, api_key: str = None):
        if username is None or len(username) == 0:
            raise StrangeworksError.authentication_error(
                message="Missing username. Use strangeworks.authenticate(username, api_key) and try again.",
            )
        if api_key is None or len(api_key) == 0:
            raise StrangeworksError.authentication_error(
                message="Missing api_key. Use strangeworks.authenticate(username, api_key) and try again.",
            )

        # Exchange api key for token to use with local SDK requests
        url = f"{self.__host}/users/{username}/token"
        self.headers["x-strangeworks-api-key"] = api_key
        res = requests.post(url, json={}, headers=self.headers, timeout=30)
        if not res.ok:
            print(
                f"Auth error: status code {res.status_code}, text {res.text} URL: {url}"
            )
            raise StrangeworksError.authentication_error(
                f"Error getting token: {res.status_code} text: {res.text} URL: {url}"
            )
        response = res.json()
        self.__auth_token = response["accessToken"]
        self.__session.auth = self.__StrangeworksAuth(token=self.__auth_token)
        self.__username = username
        self.__api_key = api_key

    def __reauthenticate(self, res: requests.Response = None, *args, **kwargs):
        if res.status_code == requests.codes.unauthorized:
            seen_before_header = "X-SW-SDK-Re-Auth"
            # we've tried once before but no luck. maybe they've changed their api_key?
            if res.request.headers.get(seen_before_header):
                raise StrangeworksError(
                    "Unable to re-authenticate your request. Utilize strangeworks.authenticate(username, api_key) with your most up to date credentials and try again."
                )
            self.authenticate(self.__username, self.__api_key)
            # self.session.send just sends the prepared request, so we must manually ensure that the new token is part of the header
            res.request.headers["Authorization"] = f"Bearer {self.__auth_token}"
            res.request.headers[seen_before_header] = True
            return self.__session.send(res.request)

    def __parse_error(self, response: requests.Response = None):
        print(f"Error from plugin: {response.text}, json: {response.json()}")
        error_payload = response.json()
        if "message" in error_payload and error_payload["message"] != "":
            raise StrangeworksError.bad_response(error_payload["message"])
        raise StrangeworksError.bad_response(
            f"Error status code: {response.status_code} text: {response.text}"
        )

    class __StrangeworksAuth(requests.auth.AuthBase):
        """
        StrangeworksAuth is used to authenticate requests to the Strangeworks
        server. Token used may be a regular Strangeworks auth token (same one
        used for API's, etc.) or a token obtained using the api key which is
        limited in scope.
        """

        def __init__(
            self,
            token: str = None,
        ):
            self.token = token

        def __call__(self, req):
            if self.token is None:
                raise StrangeworksError(
                    message="No authentication method detected. Utilize strangeworks.authenticate(username, api_key) and try again",
                )
            # utilize the authorization header with bearer token
            req.headers["Authorization"] = f"Bearer {self.token}"
            return req



from strangeworks.backend.backends import Backend, BackendService
from strangeworks.annealing.jobs import Job
from strangeworks.errors.error import StrangeworksError
from strangeworks.rest_client.rest_client import StrangeworksRestClient
from typing import Dict, Tuple, Union, List
import uuid


class Annealing:
    """
    Annealing client for direct access to annealing machines.
    """

    # TODO: replace this w/ annealing service url, currently only aws is supported
    __base_url = "/plugins/aws"

    def __init__(
        self,
        rest_client: StrangeworksRestClient = None,
        result_id: str = None,
        problem_path: str = None,
        backend: Union[str, Backend] = None,
        backends: List[Backend] = [],
    ):
        self.__rest_client = rest_client
        if self.__rest_client is None:
            raise StrangeworksError.invalid_argument(
                message="must include rest_client when initializing Annealing class, try using the default Annealing client: strangeworks.annealing...",
            )
        self.__problem_path = problem_path
        self.__result_id = result_id
        self.__backend_service = BackendService(self.__rest_client, backend, backends)

    def _new_client(self, rest_client: StrangeworksRestClient = None):
        self.__rest_client = rest_client
        self.__backend_service = BackendService(rest_client)

    def get_backends(
        self, selected_backend: Union[str, Backend] = None, pprint: bool = True
    ) -> List[Backend]:
        return self.__backend_service.get_backends(
            selected_backend, pprint, filters=lambda b: b.backend_type() == "annealing"
        )

    def select_backend(self, selected_backend: Union[str, Backend] = None) -> Backend:
        return self.__backend_service.select_backend(selected_backend)

    def upload_problem(file_path: str = ""):
        """Uploads a qubo file to Strangeworks to allow for larger problems to be defined"""
        raise StrangeworksError.not_implemented()

    def run(
        self,
        quadratic: Dict[Tuple[int, int], float] = None,
        linear: Dict[int, float] = None,
        problem_type: str = None,
        problem_path: str = None,
        shots: int = 1,
        backend: Union[str, Backend] = None,
    ) -> Job:
        """Low level annealing executor. Assumes user has already embedded problem to device target"""

        if backend:
            self.__backend_service.select_backend(backend)

        # create a new result for each job unless the user specifies a result in the client
        result_id = (
            self.__result_id if self.__result_id is not None else str(uuid.uuid4())
        )
        b = self.__backend_service.selected_backend()
        if b.backend_type() != "annealing":
            raise StrangeworksError.invalid_argument(
                f"{b.name()} is not a supported backend for annealing service"
            )
        payload = {
            "target": b.name(),
            "result_id": result_id,
            "shots": shots,
        }

        if problem_type is not None:
            payload["problem_type"] = problem_type

        # TODO: Allow uploading for problems
        problem_exists = False
        if problem_path is not None or self.__problem_path is not None:
            problem_exists = True
            raise StrangeworksError.not_implemented()

        # convert quadratic for serialization
        if quadratic is not None and linear is not None:
            nq = {}
            for v in quadratic:
                k = ",".join(map(str, v))
                nq[k] = quadratic[v]
            qb = {}
            qb["quadratic"] = nq
            qb["linear"] = linear
            payload["qubo"] = qb
            problem_exists = True

        # raise an exception if user has not provided a problem to solve
        if not problem_exists:
            raise StrangeworksError.invalid_argument(
                "must include problem to execute on annealing device"
            )

        # post to the correct plugin and return a job
        response = self.__rest_client.post(
            url=f"{self.__base_url}/run-annealing-job",
            json=payload,
            expected_response=200,
        )
        return Job.from_json(
            job=response,
            backend=self.__backend_service.selected_backend(),
            rest_client=self.__rest_client,
        )

    def __is_valid_uuid(self, value):
        if len(value) != 36:
            return False
        try:
            uuid.UUID(value)
            return True
        except ValueError:
            return False



from strangeworks.backend.backends import Backend, BackendService
from strangeworks.annealing.jobs import Job
from strangeworks.errors.error import StrangeworksError
from strangeworks.rest_client.rest_client import StrangeworksRestClient
from typing import Union, List
import uuid


class CircuitRunner:
    def __init__(
        self,
        rest_client: StrangeworksRestClient,
        result_id: str = None,
        backend: str = None,
        backends: list = [],
    ) -> None:
        self.__rest_client = rest_client
        if self.__rest_client is None:
            raise StrangeworksError.invalid_argument(
                message="must include rest_client when initializing CircuitRunner class, try using the default CircuitRunner client: strangeworks.circuit_runner...",
            )
        self.__result_id = result_id
        self.__backend_service = BackendService(rest_client, backend, backends)

    def _new_client(self, rest_client: StrangeworksRestClient = None):
        self.__rest_client = rest_client
        self.__backend_service = BackendService(rest_client)

    def get_backends(
        self, selected_backend: Union[str, Backend] = None, pprint: bool = True
    ) -> List[Backend]:
        return self.__backend_service.get_backends(
            selected_backend,
            pprint,
            filters=lambda b: b.backend_type() == "circuit_runner",
        )

    def select_backend(self, selected_backend: Union[str, Backend] = None) -> Backend:
        return self.__backend_service.select_backend(selected_backend)

    def run(
        self,
        payload: dict,
        shots: int = 1,
        backend: Union[str, Backend] = None,
        *args,
        **kwargs,
    ):
        if backend:
            self.__backend_service.select_backend(backend)

        # create a new result for each job unless the user specifies a result in the client
        result_id = (
            self.__result_id if self.__result_id is not None else str(uuid.uuid4())
        )
        b = self.__backend_service.selected_backend()
        if b.backend_type() != "circuit_runner":
            raise StrangeworksError.invalid_argument(
                f"{b.name()} is not a supported backend for circuit runner"
            )

        if "circuit" not in payload or "circuit_type" not in payload:
            raise StrangeworksError.invalid_argument(
                "Utilize one of the strangeworks extensions to convert your input"
            )
        circuit_runner_payload = {
            "target": b.name(),
            "shots": shots,
            "result_id": result_id,
        }

        circuit_runner_payload.update(kwargs)
        circuit_runner_payload.update(payload)
        
        print(circuit_runner_payload)
        
        response = self.__rest_client.post(
            url=f"{b.plugin_url()}/run-job",
            json=circuit_runner_payload,
            expected_response=200,
        )

        return Job.from_json(
            job=response,
            backend=self.__backend_service.selected_backend(),
            rest_client=self.__rest_client,
        )



from strangeworks.backend.backends import Backend
from strangeworks.errors.error import StrangeworksError
from strangeworks.rest_client.rest_client import StrangeworksRestClient
import datetime, json, time


class Job:

    COMPLETED = "completed"
    QUEUED = "queued"
    RUNNING = "running"
    FAILED = "failed"
    CANCELLED = "cancelled"
    CREATED = "created"

    __terminal_states = [COMPLETED, FAILED, CANCELLED]

    def __init__(
        self,
        id: str = None,
        remote_id: str = None,
        backend: Backend = None,
        rest_client: StrangeworksRestClient = None,
        wait_timeout: int = None,
        status: str = None,
        results: dict = {},
        slug: str = None,
    ):
        self.__id = id
        self.__remote_id = remote_id
        self.__backend = backend
        self.__rest_client = rest_client
        self.__wait_timeout = wait_timeout
        self.__status = status
        self.__results = results
        self.__slug = slug

    @classmethod
    def from_json(
        cls,
        job: dict = {},
        backend: Backend = None,
        rest_client: StrangeworksRestClient = None,
    ) -> "Job":
        id = job.get("id", "")
        remote_id = job.get("remote_id", "")
        return cls(id=id, remote_id=remote_id, backend=backend, rest_client=rest_client)

    def results(self):
        """
        waits for job completion. on error raises excetion with error.
        otherwise accepts and formats results for utilization
        """
        self.__wait_for_result()
        return self.__results

    # assumes your Job.Status is in a JOB_FINAL_STATE
    def __wait_for_result(self):
        wait = 5
        result = None
        start_time = datetime.datetime.now()
        while self.__status not in self.__terminal_states:
            response = self.__rest_client.get(f"/jobs/{self.__id}?include=result_data")
            if "status" in response:
                self.__status = response["status"]
                if self.__status == self.COMPLETED and "sdkResult" in response:
                    sdk_results = response["sdkResult"]
                    if "data" in sdk_results:
                        self.__results = json.loads(sdk_results["data"])
            time.sleep(wait)
            if self.__wait_timeout is not None and (
                (datetime.datetime.now().second - start_time.second)
                > self.__wait_timeout
            ):
                raise StrangeworksError.timeout(
                    message=f"timeout attempting to fetch results after {self.__wait_timeout}"
                )

        if self.__status == self.FAILED:
            raise StrangeworksError(
                f"Unable to get a result from an errored job {self.__slug()}"
            )

        if self.__status == self.CANCELLED:
            raise StrangeworksError(
                f"Unable to get a result from a cancelled job {self.__slug()}"
            )



"""Strangeworks SDK"""

from strangeworks.annealing.annealing import Annealing
from .auth import auth
from .VERSION import VERSION as __version__
from .client import Client


# SUBMODULES = [
#     "blueqat",
#     "braket",
#     "cirq",
#     "dwave",
#     "forest",
#     "pennylane",
#     "qiskit",
#     "tket",
# ]
# PUBLIC_METHODS = ["store_config", "print", "json", "plot", "qasm_diagram", "histogram"]

client = Client()  # instantiate a client on import by default

# strangeworks.(public method)
authenticate = client.authenticate
login = auth.Login
annealing = client.annealing
rest_client = client.rest_client
circuit_runner = client.circuit_runner

# override system print on staging and production
# ENV = os.getenv("ENV")
# if ENV in ["staging", "production"]:
#    builtins.print = client._print_log

# User's should explicity import submodules like 'import strangeworks.qiskit'.
# But if they don't, provide suggestion for importing submodules and using public
# methods. And dynamically import submodule if AttributeError would be raised.
# def __getattr__(name):
#     if name in SUBMODULES:
#         suggestion = f"You must 'import {__name__}.{name}' in your code.\n"
#         sys.stderr.write(suggestion)
#         warning = rd.error("warning", suggestion)
#         client._post_result_data(warning)
#         return importlib.import_module(f"strangeworks.{name}")
#     else:
#         suggestion = (
#             f"The strangeworks submodules available to import like "
#             f"'import strangeworks.qiskit' are: "
#             f"{SUBMODULES}.\n"
#             f"The strangeworks public methods available to use like "
#             f"'strangeworks.print(\"hello world\")' are: "
#             f"{PUBLIC_METHODS}.\n"
#         )
#         sys.stderr.write(suggestion)
#         raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from __future__ import annotations


class StrangeworksError(Exception):
    """
    This is the standard Strangeworks exception type utilized
    in both API responses to errors or from the SDK implementation
    when a known error is captured.
    """

    def __init__(
        self,
        message: str = "",
        help_page_url: str = "",
    ):
        self.message = message
        self.help_page_url = help_page_url
        super().__init__(self.message)

    @classmethod
    def authentication_error(cls, message: str = None, help_page_url: str = None):
        return cls.__builder(
            message=message,
            default_message="authentication is invalid, utilize client.authenticate() to refresh",
            help_page_url=help_page_url,
        )

    @classmethod
    def invalid_argument(cls, message: str = None, help_page_url: str = None):
        return cls.__builder(
            message=message,
            default_message="invalid argument provided",
            help_page_url=help_page_url,
        )

    @classmethod
    def not_implemented(cls, help_page_url: str = None):
        return cls.__builder(
            message="This feature is not yet implemented",
            help_page_url=help_page_url,
        )

    @classmethod
    def timeout(cls, message: str = None, help_page_url: str = None):
        return cls.__builder(
            default_message="Timeout attempting request",
            message=message,
            help_page_url=help_page_url,
        )

    @classmethod
    def bad_response(cls, message: str = "", help_page_url: str = None):
        """TODO: should replace this with many constructors based on codes and correlated help pages"""
        return cls.__builder(
            message=message,
            default_message="invalid response from server",
            help_page_url=help_page_url,
        )

    @classmethod
    def __builder(
        cls,
        message: str = None,
        help_page_url: str = None,
        default_message: str = "",
        default_help_page_url: str = "",
    ):
        sw = StrangeworksError()
        sw.message = message
        if sw.message is None:
            sw.default_message = default_message
        sw.help_page_url = help_page_url
        if help_page_url is None:
            sw.help_page_url = default_help_page_url
        return sw



from strangeworks.errors.error import StrangeworksError
from typing import List, Union


class Backend:
    __plugin_urls = {
        "aws": "/plugins/aws",
        "rigetti": "/plugins/rigetti",
        "ibm": "/plugins/ibm",
        "azure": "/plugins/azure",
    }

    def __init__(
        self,
        id: str = None,
        provider_id: str = None,
        name: str = None,
        provider_slug: str = None,
        remote_backend_id: str = None,
        plugin_url: str = None,
        backend_type: str = None,
        selector_id: str = None,
    ):
        self.__id = id
        self.__provider_id = provider_id
        self.__name = name
        self.__provider_slug = provider_slug
        self.__remote_backend_id = remote_backend_id
        self.__plugin_url = plugin_url
        self.__backend_type = backend_type
        self.__selector_id = selector_id

    @classmethod
    def from_cgw(cls, backend: dict) -> "Backend":
        id = backend.get("id", "")
        provider_id = backend.get("providerId", "")
        name = backend.get("name", "")
        provider_slug = backend.get("providerName", "")
        remote_backend_id = backend.get("remoteBackendId", "")
        backend_type = backend.get("backendType", "")                
        return Backend(
            id=id,
            provider_id=provider_id,
            name=name,
            provider_slug=provider_slug,
            remote_backend_id=remote_backend_id,
            plugin_url=cls.__plugin_urls.get(provider_slug, ""),
            backend_type=backend_type,
            selector_id=f"{provider_slug}.{name}"
        )

    def name(self) -> str:
        return self.__name

    def backend_type(self) -> str:
        return self.__backend_type

    def plugin_url(self) -> str:
        return self.__plugin_url

    def provider(self) -> str:
        return self.__provider_slug
    
    def selector_id(self) -> str:
        return self.__selector_id


class BackendService:
    def __init__(
        self,
        sw_rest_client,
        backend: Union[str, Backend] = None,
        backends: List[Backend] = [],
    ) -> None:
        self.__rest_client = sw_rest_client
        self.__backends = backends
        self.__selected_backend = BackendService.__obj_to_backend(backend)

    @classmethod
    def __obj_to_backend(cls, b: Union[str, "Backend"]) -> "Backend":
        if b and isinstance(b, str):
            b = Backend(selector_id=b)
        return b

    def __fetch_backends(
        self,
        url: str,
        selected_backend: Union[str, "Backend"],
        pprint: bool,
        filters=None,
    ):
        response = self.__rest_client.get(url)
        backends = []
        for account_slug, values in response.items():
            for value in values:
                b = value.get("backends", [])
                provider_backends = map(Backend.from_cgw, b)
                backends.extend(provider_backends)
        backends = list(filter(filters, backends))
        backends = list(filter(lambda b: not(b.provider() == "ibm" or b.provider() == "azure"), backends))
        selected_backend = BackendService.__obj_to_backend(selected_backend)
        for backend in backends:
            self.__backends.append(backend)
            if (
                selected_backend is not None
                and selected_backend.selector_id() == backend.selector_id()
            ):
                self.__selected_backend = backend
            # pprint backends if user sets option to true
            if pprint:
                print("=" * 50)
                print(f"Backend ID: {backend.selector_id()}")
                print(f"Backend provider: {backend.provider()}")

    def get_backends(
        self,
        selected_backend: Union[str, "Backend"] = None,
        pprint: bool = True,
        filters=None,
    ) -> List["Backend"]:
        # if len(self.__backends) == 0:
        self.__fetch_backends("/backends", selected_backend, pprint, filters)
        return self.__backends

    def select_backend(
        self, selected_backend: Union[str, "Backend"] = None
    ) -> "Backend":
        if not selected_backend:
            raise StrangeworksError.invalid_argument(
                "Specify a backend name or backend object"
            )
        if len(self.__backends) == 0:
            self.get_backends(pprint=False)
        selected_backend = BackendService.__obj_to_backend(selected_backend)
        selected = next(
            filter(lambda b: b.selector_id() == selected_backend.selector_id(), self.__backends), None
        )
        if not selected:
            raise StrangeworksError.invalid_argument(
                f"Backend {selected_backend.name()} not available"
            )
        self.__selected_backend = selected
        return self.__selected_backend

    def selected_backend(self) -> "Backend":
        if not self.__selected_backend:
            raise StrangeworksError.invalid_argument(
                f"No backend selected. Use select_backend to set one"
            )

        return self.__selected_backend



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strangeworks',
 'strangeworks.annealing',
 'strangeworks.auth',
 'strangeworks.backend',
 'strangeworks.circuit_runner',
 'strangeworks.errors',
 'strangeworks.rest_client']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.3,<4.0.0',
 'pylatexenc==2.7',
 'pyyaml>=5.4.1,<6.0.0',
 'requests==2.26.0']

setup_kwargs = {
    'name': 'strangeworks',
    'version': '0.2.3',
    'description': 'Strangeworks core',
    'long_description': '| ⚠️    | This SDK is currently in pre-release alpha state and subject to change. To get more info or access to test features check out the [Strangeworks Backstage Pass Program](https://strangeworks.com/backstage). |\n|---------------|:------------------------|\n\n# Strangeworks SDK\n\nPython SDK granting easy access to the Strangeworks API. For more information on using the SDK check out the [Strangeworks docs](https://docs.strangeworks.com/).',
    'author': 'Strange Devs',
    'author_email': 'stranger@strangeworks.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
